local defaultGlobals = {}
for k,_ in pairs(_G) do defaultGlobals[k] = true end

love.graphics.setBackgroundColor(0.7,0.7,0.7)

math.randomseed(os.time())

local function lerp(a,b,t)
  return a + (b - a) * t
end

local function clearTable(t)
  for k,v in pairs(t) do t[k] = nil end
end

local function doex(f,...)
  if f then return f(...) end
end

local function fullNum(n,l)
  local len = #tostring(n)
  local s = tostring(n)
  if len < l then
    for i=1,l - len do
      s = "0" .. s
    end
  end
  return s
end
local function secondFormat(sec)
  local min = math.floor(sec / 60)
  return ((min >= 60) and (math.floor(min / 60) .. ":") or "") .. fullNum(math.floor(min % 60),2) .. ":" .. fullNum(math.floor(sec % 60),2)
end

local function newArr(w,h,p)
  local toRet = {}
  for y=1,h do
    local toIns = {}
    for x=1,w do
      local c = p; if type(p) == "function" then c = p(x,y) end
      toIns[x] = c
      --table.insert(toIns,c)
    end
    table.insert(toRet,toIns)
  end
  return toRet
end

local function cloneArr(w,h,t)
  return newArr(w,h,function(x,y) return t[y][x] end)
end

--oldColor = love.graphics.setColor
--love.graphics.setColor = function(r,g,b,a) a = a or 1; oldColor(r,g,b,a) end

local feats = 3
local attribs = {"number","color","shape","shading"}

local deckId = {}

local deck = {}

local cardW,cardH = 166,109

local roundness = 15

local margin = 10

local defaultBoardH = 4
local boardW,boardH = 3,defaultBoardH

local board = {}
local displayBoard = {}

--board = newArr(boardW,boardH,boardFunc)
--displayBoard = cloneArr(board)

local selected = {}
local selectCount = 0

local score = 0

local addedRows = 0

local topBarH = 40
local font = love.graphics.newFont(topBarH-5)
local mfont = love.graphics.newFont(10)

local animTimer = 0
local animSpdDef = 1.5
local animSpd = animSpdDef
local animating = false
local animQueue = {}
local animProps = {"x","y","r","sx","sy"}

local toFill = {}

local currentSets = {}

local startTime = love.timer.getTime()

local showHint = false

local state = "game"

local winMessage = "Congratulations! You have found all possible sets."

local highlightMargin = 5

local bestTime = nil

local saveFileName = "set_save.txt"
local bestTimeName = "set_best.txt"

local newLine = "\r\n"

local function setWindowSize()
  local tw,th = boardW*(cardW+margin)+margin,boardH*(cardH+margin)+margin+topBarH
  if love.graphics.getWidth() ~= tw or love.graphics.getHeight() ~= th then
    love.window.setMode(tw,th)
  end
end

local function initCards()
  for number=1,feats do
    for color=1,feats do
      for shape=1,feats do
        for shading=1,feats do
          local c = {number=number,color=color,shape=shape,shading=shading,id=number .. color .. shape .. shading}
          c.img = love.graphics.newImage("images/card_" .. c.id .. ".png")
          deckId[c.id] = c
        end
      end
    end
  end
end
initCards()

local function generateDeck()
  deck = {}
  for id,c in pairs(deckId) do
    local i = math.random(1,#deck+1)
    table.insert(deck,i,c)
  end
end

local function inMap(x,y)
  return x >= 1 and x <= boardW and y >= 1 and y <= boardH
end

local function boardFunc(x,y)
  local c = table.remove(deck)
  c.x,c.y = x,y
  c.draw = true
  return c
end

local function mapToScreen(x,y)
  return (x-1)*(cardW+margin)+margin,(y-1)*(cardH+margin)+margin
end

local function isSet(c1,c2,c3)
  local found = true
  for _,a in ipairs(attribs) do
    if not(
      (c1[a] == c2[a] and c2[a] == c3[a]) or
      (c1[a] ~= c2[a] and c2[a] ~= c3[a] and c1[a] ~= c3[a])) then
      found = false
      break
    end
  end
  return found
end

local function checkSets()
  local sets = {}
  local checked = {}
  local function setChecked(c1,c2,c3)
    if not checked[c1] then
      checked[c1] = {}
    end
    if not checked[c1][c2] then
      checked[c1][c2] = {}
    end
    checked[c1][c2][c3] = true
  end
  for x1=1,boardW do
    for y1=1,boardH do
      for x2=1,boardW do
        for y2=1,boardH do
          for x3=1,boardW do
            for y3=1,boardH do
              local c1 = board[y1][x1]
              local c2 = board[y2][x2]
              local c3 = board[y3][x3]
              if (c1 and c2 and c3) and (c1 ~= c2 and c1 ~= c3 and c2 ~= c3) and (not checked[c1] or not checked[c1][c2] or not checked[c1][c2][c3]) and isSet(c1,c2,c3) then
                setChecked(c1,c2,c3)
                setChecked(c1,c3,c2)
                setChecked(c2,c1,c3)
                setChecked(c2,c3,c1)
                setChecked(c3,c1,c2)
                setChecked(c3,c2,c1)
                table.insert(sets,{c1,c2,c3})
              end
            end
          end
        end
      end
    end
  end
  return sets
end

local function addRow()
  boardH = boardH + 1
  addedRows = addedRows + 1
  local animToAdd = {cards={}}
  local row = {}
  for x=1,boardW do
    local c = boardFunc(x,boardH)
    c.draw = false
    table.insert(row,c)
    local fx,fy = mapToScreen(c.x,c.y)
    table.insert(animToAdd.cards,{
        x = fx,
        y1 = fy+cardH/2+topBarH,
        y2 = fy+topBarH,
        sy1 = 0,
        sy2 = 1,
        card = c
      }
    )
  end
  table.insert(board,row)
  currentSets = checkSets()
  animating = true
  animToAdd.onStart = setWindowSize
  function animToAdd.onFinish()
    local newRow = {}
    table.insert(displayBoard,newRow)
    for _,c in ipairs(row) do
      c.draw = true
      table.insert(newRow,c)
      --displayBoard[c.y][c.x] = c
    end
  end
  table.insert(animQueue,animToAdd)
end

local function clickCard(x,y)
  local c = board[y][x]
  if c then
    selected[c] = not selected[c]
    if selected[c] then
      --if we selected a card
      selectCount = selectCount + 1
      if selectCount >= 3 then
        --if we selected 3 cards
        local cardsToCheck = {}
        for c,_ in pairs(selected) do table.insert(cardsToCheck,c) end -- put selected cards inside an array, to easily unpack it
        if isSet(unpack(cardsToCheck)) then
          --if the cards we selected make a set
          animating = true
          local succCardsAnim = {cards={}}
          for _,c in ipairs(cardsToCheck) do
            displayBoard[c.y][c.x] = nil
            local fx,fy = mapToScreen(c.x,c.y)
            table.insert(succCardsAnim.cards,{
                x1 = fx,
                y1 = fy+topBarH,
                x2 = love.graphics.getWidth()/2,
                y2 = topBarH/2,
                sx1 = 1,
                sy1 = 1,
                sx2 = 0,
                sy2 = 0,
                card = c
              }
            )
          end
          table.insert(animQueue,succCardsAnim)
          score = score + 1
          local animToAdd = {cards={}}
          if addedRows > 0 then
            -- if there are added rows, fill empty spaces with the last row
            for cx=1,boardW do
              --we iterate through the cards in the last row
              local c = board[boardH][cx]
              --print(" w"..#board[boardH])
              if not selected[c] then
                --we move c to an empty space
                for ii,rc in ipairs(cardsToCheck) do
                  --c is the card we want to move, rc is the selected card we want to replace with c
                  if rc.y < boardH and not rc.replaced then
                    local fx,fy = mapToScreen(c.x,c.y)
                    board[rc.y][rc.x] = c
                    c.x,c.y = rc.x,rc.y
                    rc.replaced = true
                    local tx,ty = mapToScreen(c.x,c.y)
                    table.insert(animToAdd.cards,{
                        x1 = fx,
                        y1 = fy+topBarH,
                        x2 = tx,
                        y2 = ty+topBarH,
                        card = c
                      }
                    )
                    break
                  end
                end
              end
            end
            animToAdd.onStart = function() for _,c in ipairs(animToAdd.cards) do c.card.draw = false end end
            animToAdd.onFinish = function() setWindowSize(); table.remove(displayBoard) end
            --remove the last row
            table.remove(board)
            boardH = boardH - 1
            addedRows = addedRows - 1
          elseif #deck >= 3 then
            --replace selected cards with new cards from the deck
            for _,c in ipairs(cardsToCheck) do
              local nc = boardFunc(c.x,c.y)
              nc.draw = false
              board[c.y][c.x] = nc
              local fx,fy = mapToScreen(c.x,c.y)
              table.insert(animToAdd.cards,{
                  x = fx,
                  y1 = fy+cardH/2+topBarH,
                  y2 = fy+topBarH,
                  sy1 = 0,
                  sy2 = 1,
                  card = nc
                }
              )
            end
          else
            for c,_ in pairs(selected) do board[c.y][c.x] = nil end
          end
          if #animToAdd.cards > 0 then
            local oldFunc = animToAdd.onFinish
            function animToAdd.onFinish()
              doex(oldFunc)
              for _,c in ipairs(animToAdd.cards) do
                c.card.draw = true
                displayBoard[c.card.y][c.card.x] = c.card
              end
            end
            table.insert(animQueue,animToAdd)
          end
          currentSets = checkSets()
          if #currentSets == 0 then
            if #deck >= 3 then
              repeat
                addRow()
                currentSets = checkSets()
              until #currentSets > 0
            else
              if not bestTime or time < bestTime then
                bestTime = time
                love.filesystem.write(bestTimeName,tostring(bestTime))
              end
              function animToAdd.onStart()
                animSpd = 100
              end
              function animToAdd.onFinish()
                state = "win"
              end
              table.insert(animQueue,animToAdd)
            end
          end
        end
        --reset selection
        clearTable(selected)
        selectCount = 0
      end
    else
      --if we deselected a card
      selectCount = selectCount - 1
      selected[c] = nil
    end
  end
end

--currentSets = checkSets()

local function updateAnim(a)
  if not a.started then
    doex(a.onStart)
    a.started = true
  end
  if a.cards then
    for _,c in ipairs(a.cards) do
      for _,p in ipairs(animProps) do
        if c[p..1] then c[p] = lerp(c[p..1],c[p..2],animTimer) end
      end
    end
  end
  doex(a.update,a)
end

local function cheat()
  for _,c in ipairs(checkSets()[1]) do
    clickCard(c.x,c.y)
  end
end

function love.load(arg)
  startTime = love.timer.getTime()
  local c = love.filesystem.read(bestTimeName)
  bestTime = tonumber(c)
  local save = love.filesystem.read(saveFileName)
  if save then
    deck = {}
    local lines = {}
    for l in string.gmatch(save,"[^" .. newLine .. "]+") do table.insert(lines,l) end
    if lines[1] ~= "0" then
      for id in string.gmatch(lines[1],"[^%,]+") do
        table.insert(deck,deckId[id])
      end
    end
    board = {}
    local x,y = 0,0
    for l in string.gmatch(lines[2],"[^%|]+") do
      x = 0
      y = y + 1
      local toIns = {}
      for id in string.gmatch(l,"[^%,]+") do
        --print(id)
        x = x + 1
        local c = deckId[id]
        --print(x,y,c)
        if c then
          c.draw = true
          c.x,c.y = x,y
        end
        toIns[x] = c
        --table.insert(toIns,c)
      end
      table.insert(board,toIns)
    end
    boardH = #board
    addedRows = boardH - defaultBoardH
    score = lines[3]
    startTime = startTime - lines[4]
  else
    --shuffle(deck)
    repeat
      generateDeck()
      board = newArr(boardW,boardH,boardFunc)
    until #checkSets() > 0
  end
  displayBoard = cloneArr(boardW,boardH,board)
  setWindowSize()
  currentSets = checkSets()
  for k,_ in pairs(_G) do
    if not defaultGlobals[k] then print(k) end
  end
end

function love.mousepressed(x,y,b)
  if b == 1 then
    if y <= topBarH then
      if x >= love.graphics.getWidth()-topBarH then
        local answer = love.window.showMessageBox("Start new game","Start new game?",{"Yes","No"})
        if answer == 1 then
          animSpd = animSpdDef
          highlightCard = nil
          animating = false
          animQueue = {}
          animTimer = 0
          state = "game"
          score = 0
          addedRows = 0
          startTime = love.timer.getTime()
          boardH = defaultBoardH
          repeat
            generateDeck()
            board = newArr(boardW,boardH,boardFunc)
            currentSets = checkSets()
          until #currentSets > 0
          displayBoard = cloneArr(boardW,boardH,board)
          setWindowSize()
        end
      elseif x >= love.graphics.getWidth()-topBarH*2 and state == "game" then
        local answer = love.window.showMessageBox("Show hint","Show a hint?",{"Yes","No"})
        if answer == 1 then
          highlightCard = currentSets[math.random(#currentSets)][math.random(3)]
        end
      end
    elseif state == "game" and not animating then
      highlightCard = nil
      local tx,ty = math.ceil(x/(cardW+margin)),math.ceil((y-topBarH)/(cardH+margin))
      if inMap(tx,ty) then
        clickCard(tx,ty)
      end
    end
  --[[elseif b == 3 then
    cheat()]]
  end
end

function love.update(dt)
  if state == "game" then
    time = love.timer.getTime()-startTime -- + 60*59 + 40
    if animating then
      local anim = animQueue[1]
      animTimer = animTimer + animSpd * dt
      updateAnim(anim)
      if animTimer >= 1 then
        animTimer = 0
        local rem = table.remove(animQueue,1)
        doex(rem.onFinish)
        if #animQueue <= 0 then
          animating = false
        else
          updateAnim(animQueue[1])
        end
      end
    end
    --showHint = love.mouse.getY() <= topBarH and love.mouse.isDown(1)
  end
end

function love.draw()
  love.graphics.setColor(0,0,0,0.7)
  love.graphics.rectangle("fill",0,0,love.graphics.getWidth(),topBarH)
  love.graphics.setColor(1,1,1)
  love.graphics.setFont(font)
  love.graphics.printf(score,0,0,love.graphics.getWidth(),"center")
  if state == "win" then
    love.graphics.printf({{0,1,0},secondFormat(time),{1,1,0}," "..secondFormat(bestTime)},0,0,love.graphics.getWidth(),"left")
  else
    love.graphics.printf(secondFormat(time),0,0,love.graphics.getWidth(),"left")
  end
  if state == "game" then
    love.graphics.setColor(0.9,0.9,0.9)
    love.graphics.rectangle("fill",love.graphics.getWidth()-topBarH*2,0,topBarH,topBarH)
    love.graphics.setColor(0,0,0)
    love.graphics.printf("?",love.graphics.getWidth()-topBarH*2,0,topBarH,"center")
  end
  love.graphics.push()
  love.graphics.translate(love.graphics.getWidth()-topBarH/2,topBarH/2)
  if state == "win" then
    local sc = (math.sin(love.timer.getTime()*5)+1)/5
    love.graphics.scale(1+sc,1+sc)
  end
  love.graphics.setColor(1,1,1)
  love.graphics.rectangle("fill",-topBarH/2,-topBarH/2,topBarH,topBarH)
  love.graphics.setColor(0,0,0)
  love.graphics.setLineWidth(2)
  for i=1,8 do
    local a = i/8*math.pi*2-- + ((state == "win") and (love.timer.getTime()/2) or 0)
    local dx,dy = math.cos(a),math.sin(a)
    love.graphics.line(dx*topBarH/4,dy*topBarH/4,dx*topBarH/2.5,dy*topBarH/2.5)
  end
  love.graphics.pop()
  ----------
  if state == "game" then
    love.graphics.push()
    love.graphics.translate(0,topBarH)
    for x=1,boardW do
      for y=1,#displayBoard do
        local c = displayBoard[y][x]
        if c and c.draw then
          local dx,dy = mapToScreen(x,y)
          if selected[c] then
            love.graphics.setColor(0,0,0,0.7)
            love.graphics.rectangle("fill",dx,dy,cardW,cardH,roundness,roundness)
            dx,dy = dx-margin,dy-margin
            love.graphics.setLineWidth(5)
            love.graphics.setColor(0,1,1)
            love.graphics.rectangle("line",dx,dy,cardW,cardH,roundness,roundness)
          end
          love.graphics.setColor(1,1,1)
          if showHint and #currentSets > 0 then
            --cycle through all available sets
            for _,cc in ipairs(currentSets[math.ceil(love.timer.getTime())%#currentSets+1]) do
              if c == cc then
                love.graphics.setColor(1,0.5,0.5)
              end
            end
          end
          love.graphics.draw(c.img,dx,dy)
        end
      end
    end
    if highlightCard then
      local dx,dy = mapToScreen(highlightCard.x,highlightCard.y)
      local m = highlightMargin + math.sin(love.timer.getTime()*5) * 5
      love.graphics.setColor(1,0,1)
      love.graphics.setLineWidth(4)
      love.graphics.rectangle("line",dx-m,dy-m,cardW+m*2,cardH+m*2,roundness,roundness)
    end
    love.graphics.pop()
    if animating then
      local anim = animQueue[1]
      love.graphics.setColor(1,1,1)
      if anim.cards then
        for _,a in ipairs(anim.cards) do
          love.graphics.draw(a.card.img,a.x,a.y,a.r,a.sx,a.sy)
        end
      end
      doex(anim.draw,anim)
    end
  else
    love.graphics.setColor(0,0,0)
    love.graphics.setFont(font)
    love.graphics.printf(winMessage,0,love.graphics.getHeight()/2-font:getHeight()/2,love.graphics.getWidth(),"center")
  end
  ----------
end

function love.quit()
  if state == "game" then
    local data = ""
    if #deck > 0 then
      for _,c in ipairs(deck) do
        data = data .. c.id .. ","
      end
    else
      data = data .. "0"
    end
    data = data .. newLine
    for y=1,boardH do
      for x=1,boardW do
        data = data .. (board[y][x] and board[y][x].id or "0") .. ","
      end
      data = data .. "|"
    end
    data = data .. newLine .. score
    data = data .. newLine .. time
    love.filesystem.write(saveFileName,data)
  end
end